"""Provide fixtures for test suite."""

from pathlib import Path
from typing import Tuple

import numpy as np
import pytest


@pytest.fixture(scope="package")
def ds_path(request) -> Path:
    """Provide the path to the data set folder."""
    curr = Path(request.fspath).parents[0]
    return Path(curr).joinpath("datasets").absolute()


@pytest.fixture(scope="package")
def stephan_ci_data(
    ds_path,
) -> Tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
    np.ndarray,
]:
    """Provide confidence interval data from Stephan Macro for testing.

    The data and results are stored in the datasets folder.

    :param ds_path: Pytest fixture to provide the path to the data.

    :return: xdat, sigx, ydat, sigy, rho, x_ci, y_ci_min, y_ci_max
    """
    data_fname = ds_path.joinpath("ci_calc_data.csv")
    results_fname = ds_path.joinpath("ci_calc_results.csv")

    data = np.loadtxt(data_fname, skiprows=1, delimiter=",")
    results = np.loadtxt(results_fname, skiprows=1, delimiter=",")

    return (
        data[:, 0],
        data[:, 1],
        data[:, 2],
        data[:, 3],
        data[:, 4],
        results[:, 0],
        results[:, 1],
        results[:, 2],
    )



"""Test properties, error messages, etc. on Mahon Linear regression."""

import numpy as np
import pytest

from ceresfit import LinReg
from .test_ceresfit_linreg_datasets import read_dataset


def test_linreg_fixpoint_value_error():
    """Raise ValueError if fix point is of bad shape."""
    some_arr = np.array([1, 2])
    with pytest.raises(ValueError):
        _ = LinReg(some_arr, some_arr, some_arr, some_arr, fixpt=42, autocalc=False)


def test_linreg_keywords(ds_path):
    """Ensure keywords are accepted by routine."""
    case = "set1.csv"
    ds = ds_path.joinpath(case)
    xdat, xunc, ydat, yunc, rho, _, params_exp = read_dataset(ds)

    reg_limit = 42
    iter_max = 13
    kwargs = {"regression_limit": reg_limit, "iter_max": iter_max}

    reg = LinReg(xdat, xunc, ydat, yunc, rho=rho, autocalc=False, **kwargs)

    assert reg.reg_limit == reg_limit
    assert reg.iter_max == iter_max


def test_linreg_iteration_warning(ds_path):
    """Test that iteration throws a warning if runs out of max_iter."""
    case = "set1.csv"
    ds = ds_path.joinpath(case)
    xdat, xunc, ydat, yunc, rho, _, params_exp = read_dataset(ds)
    iter_max = 1

    with pytest.warns(UserWarning):
        _ = LinReg(xdat, xunc, ydat, yunc, rho=rho, iter_max=iter_max)


def test_linreg_properties(ds_path):
    """Test that properties return the correct parameters."""
    case = "set1.csv"
    ds = ds_path.joinpath(case)
    xdat, xunc, ydat, yunc, rho, _, params_exp = read_dataset(ds)

    ind_params = 0  # index of which parameters for given case

    reg = LinReg(xdat, xunc, ydat, yunc, rho=rho)

    assert params_exp[ind_params][0:2] == pytest.approx(reg.slope, abs=1e-6)
    assert params_exp[ind_params][2:4] == pytest.approx(reg.intercept, abs=1e-6)
    assert params_exp[ind_params][4] == pytest.approx(reg.mswd, abs=1e-3)


def test_linreg_mswd_ci():
    """Confidence intervals for MSWD."""
    dof = 9
    mswd_ci_exp = 0.300043277775595, 2.11364086651574
    some_arr = np.array([1, 2])

    reg = LinReg(some_arr, some_arr, some_arr, some_arr, auto=False)
    reg._dof = dof

    mswd_ci_rec = reg.mswd_ci()
    assert mswd_ci_exp == pytest.approx(mswd_ci_rec)


def test_linreg_regression_line():
    """Get a regression line for the specified values."""
    some_arr = np.array([1, 2])

    reg = LinReg(some_arr, some_arr, some_arr, some_arr, auto=False)
    reg._slope = 3.14
    reg._intercept = 12.0

    xdat_exp = np.array([1, 2])
    ydat_exp = xdat_exp * reg._slope + reg._intercept

    xdat_rec, ydat_rec = reg.regression_line()

    assert xdat_rec == pytest.approx(xdat_exp)
    assert ydat_rec == pytest.approx(ydat_exp)



"""Linear Regression according to Mahon (1996) - corrected."""

from typing import Tuple, Union
import warnings

import numpy as np
from numpy.polynomial import Polynomial
import scipy.stats as stats
from scipy.stats.distributions import chi2


class LinReg:
    """Linear regression using (corrected) Mahon (1996) prescription.

    Example:
        >>> import numpy as np
        >>> from ceresfit import LinReg
        >>> # some data
        >>> xdata = np.array([1, 2, 3.1, 4.9])
        >>> ydata = np.array([1.1, 1.9, 3, 5.5])
        >>> # some uncertainty and correlation
        >>> xunc = 0.05 * xdata
        >>> yunc = 0.073 * ydata
        >>> rho = np.zeros_like(xdata) + 0.5
        >>> # do regression
        >>> my_reg = LinReg(xdata, xunc, ydata, yunc, rho)
        >>> # print out the parameters and their uncertainties
        >>> my_reg.slope
        (0.9983613298400896, 0.06844666435449052)
        >>> my_reg.intercept
        (0.05545398718611372, 0.11812746374874884)
        >>> my_reg.mswd
        2.5105964767071143
    """

    def __init__(
        self,
        xdat: np.ndarray,
        sigx: np.ndarray,
        ydat: np.ndarray,
        sigy: np.ndarray,
        rho: Union[float, np.ndarray] = None,
        fixpt: np.ndarray = None,
        autocalc=True,
        **kwargs,
    ):
        """Initialize the class.

        :param xdat: X data.
        :param ydat: Y data.
        :param sigx: 1 sigma uncertainty of x data.
        :param sigy: 1 sigma uncertainty of y data.
        :param rho: Correlation between x and y data, defaults to no correlation.
        :param fixpt: Fixed point through which regression needs to go.
        :param autocalc: Automatically calculate the regression and print params.
        :param kwargs: Additional keyword arguments:
            "iter_max": Maximum iteration limit for slope (default 1e6)
            "reg_limit": Regression limit for slope (default: 1e-6)

        :raises ValueError: Fix point is of the wrong shape.
        """
        self.xdat = np.array(xdat)
        self.sigx = np.array(sigx)
        self.ydat = np.array(ydat)
        self.sigy = np.array(sigy)
        if rho is None:
            self.rho = np.zeros_like(self.xdat)
        else:
            self.rho = np.array(rho)

        # calculate correlated uncertainty sigxy
        self.sigxy = self.rho * self.sigx * self.sigy

        if fixpt is not None:
            fixpt = np.array(fixpt)
            if fixpt.shape != (2,):
                raise ValueError("Fix point must be of the form [x_fix, yfix].")
        self.fix_pt = fixpt

        # Initialize the parameters to be returned
        self._dof = None
        self._slope = None
        self._slope_unc = None
        self._intercept = None
        self._intercept_unc = None
        self._chi_squared = None
        self._mswd = None

        # keyword arguments
        if "regression_limit" in kwargs:
            self.reg_limit = kwargs["regression_limit"]
        else:
            self.reg_limit = 1e-12
        if "iter_max" in kwargs:
            self.iter_max = kwargs["iter_max"]
        else:
            self.iter_max = 1e6

        # helper variables
        self.xbar = None
        self.ybar = None
        self.weights = None

        if autocalc:
            self.calculate()

    @property
    def chi_squared(self) -> float:
        """Return chi_squared of the regression."""
        return self._chi_squared

    @property
    def dof(self) -> int:
        """Return the degrees of freedom of the system."""
        return self._dof

    @property
    def intercept(self) -> Tuple[float, float]:
        """Return intercept and its 1 sigma uncertainty."""
        return self._intercept, self._intercept_unc

    @property
    def mswd(self) -> float:
        """Return MSWD of the regression."""
        return self._mswd

    @property
    def parameters(self) -> np.ndarray:
        """Return all parameters of the linear regression.

        :return: slope, slope_uncertainty, intercept, intercept_uncertainty, MSWD
        """
        return np.array(
            [
                self._slope,
                self._slope_unc,
                self._intercept,
                self._intercept_unc,
                self._mswd,
            ]
        )

    @property
    def slope(self) -> Tuple[float, float]:
        """Return slope and its 1 sigma uncertainty."""
        return self._slope, self._slope_unc

    def calculate(self):
        """Do the linear regression and save the parameters in the class variables."""
        self.slope_initial_guess()
        self.slope_calculation()
        self.intercept_calculation()
        self.unc_calculation()
        self.goodness_of_fit()

    def confidence_intervals(
        self,
        p_conf: float = 0.95,
        xrange: np.ndarray = None,
        bins: int = 100,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Calculate the confidence intervals and return them.

        :param p_conf: Confidence interval, default 95% (``p_conf=0.95``).
        :param xrange: Range from where to where to calculate CI. Default ``None``.
            If None, defaults to min and max of x values that are defined.
        :param bins: Number of bins to calculate, defaults to 100.

        :return: Data to plot CI: xdata, y_ci_minimum, y_ci_maximum
        """
        xax_ci, yax_ub_min, yax_ub_max = self.uncertainty_band(
            sigma=1, xrange=xrange, bins=bins
        )

        yax_ci = np.abs(yax_ub_max - yax_ub_min) / 2.0

        # now create the confidence interval that we need for double tailed distribution
        zfac = stats.t.ppf(1 - (1 - p_conf) / 2.0, len(self.xdat) - 2)
        yax_ci *= zfac

        yax_ci_min = xax_ci * self.slope[0] + self.intercept[0] - yax_ci
        yax_ci_max = xax_ci * self.slope[0] + self.intercept[0] + yax_ci

        return xax_ci, yax_ci_min, yax_ci_max

    def goodness_of_fit(self):
        """Calculate goodness of fit parameters chi-squared and MSWD."""
        chi_sq = np.sum(
            self.weights * (self.ydat - self._slope * self.xdat - self._intercept) ** 2
        )
        self._chi_squared = chi_sq

        dof = len(self.xdat) - 2 if self.fix_pt is None else len(self.xdat) - 1
        self._dof = dof
        self._mswd = chi_sq / dof

    def intercept_calculation(self):
        """Calculate the intercept."""
        self._intercept = self.ybar - self._slope * self.xbar

    def mswd_ci(self, p_conf=0.95) -> Tuple[float, float]:
        """Calculate confidence interval of MSWD value given degrees of freedom of.

        :param p_conf: Confidence level, defaults to 0.95 (95%)

        :return: Confidence interval for the MSWD inside this interval.
        """
        dof = self.dof
        right_tail = 1 - (1 - p_conf) / 2
        left_tail = 1 - right_tail
        mswd_lower = chi2.ppf(left_tail, dof) / dof
        mswd_upper = chi2.ppf(right_tail, dof) / dof
        return mswd_lower, mswd_upper

    def slope_calculation(self):
        """Iterate the slope until it fits."""

        def calc_weights(b: float):
            """Calculate weights and return them.

            :param b: Slope

            :return: Weights
            """
            return 1 / (self.sigy**2 + b**2 * self.sigx**2 - 2 * b * self.sigxy)

        def calc_xbar(weights: np.ndarray):
            """Calculate x bar and return it.

            :param weights: Weights.

            :return: X bar.
            """
            if self.fix_pt is None:
                return np.sum(weights * self.xdat) / np.sum(weights)
            else:
                return self.fix_pt[0]

        def calc_ybar(weights: np.ndarray):
            """Calculate y bar and return it.

            :param weights: Weights.

            :return: Y bar.
            """
            if self.fix_pt is None:
                return np.sum(weights * self.ydat) / np.sum(weights)
            else:
                return self.fix_pt[1]

        def iterate_b(b_old):
            """Do one iteration step with the slope and return the new value."""
            b = b_old
            weights = calc_weights(b)
            u_all = self.xdat - calc_xbar(weights)
            v_all = self.ydat - calc_ybar(weights)
            b_new = np.sum(
                weights**2
                * v_all
                * (
                    u_all * self.sigy**2
                    + b * v_all * self.sigx**2
                    - v_all * self.sigxy
                )
            ) / np.sum(
                weights**2
                * u_all
                * (
                    u_all * self.sigy**2
                    + b * v_all * self.sigx**2
                    - b * u_all * self.sigxy
                )
            )
            return b_new

        # iterate until solution is found
        iter_cnt = 0
        b_old = self._slope
        b_new = iterate_b(b_old)
        while np.abs(b_old - b_new) > self.reg_limit and iter_cnt < self.iter_max:
            b_old = b_new
            b_new = iterate_b(b_old)
            iter_cnt += 1

        if iter_cnt >= self.iter_max:
            warnings.warn(
                f"Iteration count for slope optimization hit the limt at "
                f"{self.iter_max}. The current difference between the old and new "
                f"slope is {np.abs(b_old - b_new)}",
                stacklevel=1,
            )

        self._slope = b_new
        self.weights = calc_weights(b_new)
        self.xbar = calc_xbar(self.weights)
        self.ybar = calc_ybar(self.weights)

    def slope_initial_guess(self):
        """Calculate an initial guess of the slope without uncertainties and save it."""
        polyfit = Polynomial.fit(self.xdat, self.ydat, deg=1)
        self._slope = polyfit.convert().coef[1]

    def uncertainty_band(
        self,
        sigma=1,
        xrange: np.ndarray = None,
        bins: int = 100,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Calculate the uncertainty bands and return them.

        :param sigma: How many sigma should the band be? Default: 1
        :param xrange: Range from where to where to band. Default ``None``.
            If None, defaults to min and max of x values that are defined.
        :param bins: Number of bins to calculate, defaults to 100.

        :return: Data to plot band: xdata, y_ub_minimum, y_ub_maximum
        """
        if xrange is None:
            xax_ci = np.linspace(np.min(self.xdat), np.max(self.xdat), bins)
        else:
            xax_ci = np.linspace(xrange[0], xrange[1], bins)

        xdat_save = self.xdat.copy()

        yax_ub = np.zeros(bins)
        for it, deltax in enumerate(xax_ci):
            self.xdat = xdat_save - deltax
            self.slope_initial_guess()
            self.slope_calculation()
            self.intercept_calculation()
            self.unc_calculation()
            yax_ub[it] = self.intercept[1]

        # mulitply with sigma
        yax_ub *= sigma

        # reset all calculations
        self.xdat = xdat_save
        self.calculate()

        yax_ub_min = xax_ci * self.slope[0] + self.intercept[0] - yax_ub
        yax_ub_max = xax_ci * self.slope[0] + self.intercept[0] + yax_ub

        return xax_ci, yax_ub_min, yax_ub_max

    def unc_calculation(self):
        """Uncertainty calculation.

        This method is called by the ``calculate`` method and simply calls the
        ``unc_calculation_no_fix_point`` or ``unc_calculation_fix_point`` method.
        """
        if self.fix_pt is None:
            self.unc_calculation_no_fix_point()
        else:
            self.unc_calculation_fix_point()

    def unc_calculation_no_fix_point(self):
        """Calculate uncertainties for slope and intercept with no fixed point."""
        # helper variables
        sigx = self.sigx
        sigy = self.sigy
        sigxy = self.sigxy
        b = self._slope
        weights = self.weights
        xbar = self.xbar
        ybar = self.ybar
        u_all = self.xdat - xbar
        v_all = self.ydat - ybar

        sum_weights = np.sum(weights)

        # sums over j in equation 60, last two lines
        sum_j_u = np.sum(weights**2 * u_all * (sigxy - b * sigx**2))
        sum_j_v = np.sum(weights**2 * v_all * (sigxy - b * sigx**2))

        # d(theta) / db
        dthdb = (
            np.sum(
                weights**2
                * (
                    2 * b * (u_all * v_all * sigx**2 - u_all**2 * sigxy)
                    + (u_all**2 * sigy**2 - v_all**2 * sigx**2)
                )
            )
            + 4
            * np.sum(
                weights**3
                * (sigxy - b * sigx**2)
                * (
                    b**2 * (u_all * v_all * sigx**2 - u_all**2 * sigxy)
                    + b * (u_all**2 * sigy**2 - v_all**2 * sigx**2)
                    - (u_all * v_all * sigy**2 - v_all**2 * sigxy)
                )
            )
            + 2
            * np.sum(
                weights**2
                * (
                    -(b**2) * u_all * sigx**2
                    + 2 * b * v_all * sigx**2
                    + u_all * sigy**2
                    - 2 * v_all * sigxy
                )
            )
            * sum_j_v
            / sum_weights
            + 2
            * np.sum(
                weights**2
                * (
                    -(b**2) * v_all * sigx**2
                    + 2 * b**2 * u_all * sigxy
                    - 2 * b * u_all * sigy**2
                    + v_all * sigy**2
                )
            )
            * sum_j_u
            / sum_weights
        )

        def calc_dtheta_dxi(it: int):
            """Calculate partial derivative d(theta)/dxi.

            :param it: Index where the $i$ is at.

            :return: dtheta/dxi
            """
            sum_all = 0.0
            for jt, wj in enumerate(weights):
                kron = kron_delta(it, jt)
                sum_all += (
                    wj**2
                    * (kron - weights[it] / sum_weights)
                    * (
                        b**2 * v_all[jt] * sigx[jt] ** 2
                        - b**2 * 2 * u_all[jt] * sigxy[jt]
                        + 2 * b * u_all[jt] * sigy[jt] ** 2
                        - v_all[jt] * sigy[jt] ** 2
                    )
                )
            return sum_all

        def calc_dtheta_dyi(it: int):
            """Calculate partial derivative d(theta)/dyi.

            :param it: Index where the $i$ is at.

            :return: dtheta/dyi
            """
            sum_all = 0.0
            for jt, wj in enumerate(weights):
                kron = kron_delta(it, jt)
                sum_all += (
                    wj**2
                    * (kron - weights[it] / sum_weights)
                    * (
                        b**2 * u_all[jt] * sigx[jt] ** 2
                        - 2 * b * v_all[jt] * sigx[jt] ** 2
                        - u_all[jt] * sigy[jt] ** 2
                        + 2 * v_all[jt] * sigxy[jt]
                    )
                )
            return sum_all

        def calc_da_dxi(it: int):
            """Calculate partial derivative da/dxi.

            This routine is only used when no fixed point is selected.

            :param it: Index where the $i$ is at.

            :return: da/dxi
            """
            return -b * weights[it] / sum_weights - xbar * calc_dtheta_dxi(it) / dthdb

        def calc_da_dyi(it: int):
            """Calculate partial derivative da/dyi.

            This routine is only used when no fixed point is selected.

            :param it: Index where the $i$ is at.

            :return: da/dyi
            """
            return weights[it] / sum_weights - xbar * calc_dtheta_dyi(it) / dthdb

        # calculate uncertainty for slope
        sigb_sq = 0.0
        for it, sigxi in enumerate(sigx):
            sigyi = sigy[it]
            sigxyi = sigxy[it]
            dtheta_dxi = calc_dtheta_dxi(it)
            dtheta_dyi = calc_dtheta_dyi(it)
            sigb_sq += (
                dtheta_dxi**2 * sigxi**2
                + dtheta_dyi**2 * sigyi**2
                + 2 * sigxyi * dtheta_dxi * dtheta_dyi
            )
        sigb_sq /= dthdb**2
        self._slope_unc = np.sqrt(sigb_sq)

        siga_sq = 0.0
        for it, sigxi in enumerate(sigx):
            sigyi = sigy[it]
            sigxyi = sigxy[it]
            da_dxi = calc_da_dxi(it)
            da_dyi = calc_da_dyi(it)
            siga_sq += (
                da_dxi**2 * sigxi**2
                + da_dyi**2 * sigyi**2
                + 2 * sigxyi * da_dxi * da_dyi
            )
        self._intercept_unc = np.sqrt(siga_sq)

    def unc_calculation_fix_point(self):
        """Calculate uncertainties for slope and intercept with fixed point."""
        # helper variables
        sigx = self.sigx
        sigy = self.sigy
        sigxy = self.sigxy
        b = self._slope
        weights = self.weights
        xbar = self.xbar
        ybar = self.ybar
        u_all = self.xdat - xbar
        v_all = self.ydat - ybar

        # d(theta) / db
        dthdb = np.sum(
            weights**2
            * (
                2 * b * (u_all * v_all * sigx**2 - u_all**2 * sigxy)
                + (u_all**2 * sigy**2 - v_all**2 * sigx**2)
            )
        ) + 4 * np.sum(
            weights**3
            * (sigxy - b * sigx**2)
            * (
                b**2 * (u_all * v_all * sigx**2 - u_all**2 * sigxy)
                + b * (u_all**2 * sigy**2 - v_all**2 * sigx**2)
                - (u_all * v_all * sigy**2 - v_all**2 * sigxy)
            )
        )

        def calc_dtheta_dxi(it: int):
            """Calculate partial derivative d(theta)/dxi.

            :param it: Index where the $i$ is at.

            :return: dtheta/dxi
            """
            return weights[it] ** 2 * (
                b**2 * v_all[it] * sigx[it] ** 2
                - b**2 * 2 * u_all[it] * sigxy[it]
                + 2 * b * u_all[it] * sigy[it] ** 2
                - v_all[it] * sigy[it] ** 2
            )

        def calc_dtheta_dyi(it: int):
            """Calculate partial derivative d(theta)/dyi.

            :param it: Index where the $i$ is at.

            :return: dtheta/dyi
            """
            return weights[it] ** 2 * (
                b**2 * u_all[it] * sigx[it] ** 2
                - 2 * b * v_all[it] * sigx[it] ** 2
                - u_all[it] * sigy[it] ** 2
                + 2 * v_all[it] * sigxy[it]
            )

        # calculate uncertainty for slope
        sigb_sq = 0.0
        for it, sigxi in enumerate(sigx):
            sigyi = sigy[it]
            sigxyi = sigxy[it]
            dtheta_dxi = calc_dtheta_dxi(it)
            dtheta_dyi = calc_dtheta_dyi(it)
            sigb_sq += (
                dtheta_dxi**2 * sigxi**2
                + dtheta_dyi**2 * sigyi**2
                + 2 * sigxyi * dtheta_dxi * dtheta_dyi
            )
        sigb_sq /= dthdb**2
        self._slope_unc = np.sqrt(sigb_sq)

        siga_sq = self.fix_pt[0] ** 2 * sigb_sq
        print(siga_sq)
        self._intercept_unc = np.sqrt(siga_sq)

    def regression_line(
        self, xrange: np.ndarray = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Return xdata and ydata for the calculated regression line for plotting.

        :param xrange: Range for which to calculate the regression. Default: ``None``.
            If None, defaults to min and max of x values that are defined.

        :return: xdata to plot, ydata to plot regression line.
        """
        if xrange is None:
            xrange = np.array([self.xdat.min(), self.xdat.max()])

        ydat = self._slope * xrange + self._intercept
        return xrange, ydat


def kron_delta(
    ind1: Union[int, np.ndarray], ind2: Union[int, np.ndarray]
) -> Union[int, np.ndarray]:
    """Calculate Kronecker-delta for variables i,j.

    Compare two indexes and return 0 if the same, otherwise 1. If an ndarray is given,
    return an ndarray comparing each index individually.

    :param ind1: Index(es)
    :param ind2: Index(es)

    :return: 1 if ind 1 is identical to ind2, otherwise 0

    :raises ValueError: The input indexes have different shape.
    """
    if np.shape(ind1) != np.shape(ind2):
        raise ValueError("The inputs must have the same shape.")

    if np.shape(ind1) == ():  # don't have arrays
        return 1 if ind1 == ind2 else 0
    else:
        ret_arr = np.zeros_like(ind1)
        ret_arr[np.where(ind1 == ind2)] = 1
        return ret_arr



"""Test uncertainty calculations."""

import numpy as np
import pytest

from ceresfit import unc_calc as uc


def test_error_bar_angle_length_single():
    """Calculate error bar angle and length for given correlation (single value)."""
    sigx = 1
    sigy = 2
    rho = 0.75

    angle_exp = np.array([np.deg2rad(67.5), np.deg2rad(-22.5)])
    length_exp = np.array([np.sqrt(2.5 + np.sqrt(4.5)), np.sqrt(2.5 - np.sqrt(4.5))])

    angle_rec, length_rec = uc.error_bar_angle_length(sigx, sigy, rho)

    assert angle_exp == pytest.approx(angle_rec)
    assert length_exp == pytest.approx(length_rec)


@pytest.mark.parametrize("rho", [0.75, np.array([0.75, 0.75])])
def test_error_bar_angle_length_multi(rho):
    """Calculate error bar angle and length for given correlation (multiple values)."""
    sigx = np.array([1, 2])
    sigy = np.array([2, 1])

    length_exp = np.array(
        [
            np.sqrt(2.5 + np.sqrt(4.5)),
            np.sqrt(2.5 - np.sqrt(4.5)),
            np.sqrt(2.5 + np.sqrt(4.5)),
            np.sqrt(2.5 - np.sqrt(4.5)),
        ]
    )

    sigxy = rho * sigx * sigy
    angle_exp = np.array(
        [
            np.arctan((length_exp[:2] ** 2 - sigx[0] ** 2) / sigxy[0]),
            np.arctan((length_exp[2:] ** 2 - sigx[1] ** 2) / sigxy[1]),
        ]
    ).reshape(4)

    angle_rec, length_rec = uc.error_bar_angle_length(sigx, sigy, rho)

    assert angle_exp == pytest.approx(angle_rec)
    assert length_exp == pytest.approx(length_rec)


def test_error_bar_position_single():
    """Test error bar position with Thomas' Macro value."""
    xdat = 49.5263059304119
    sigx = 116.127138038472
    ydat = 94.8320931109692
    sigy = 137.018340778937
    rho = 0.429886405472074

    xpos_exp = np.array(
        [[-37.0849520308128, 136.137563891636], [-27.8300266053324, 126.882638466156]]
    )
    ypos_exp = np.array(
        [[-31.507663087177, 221.171849309115], [147.863137920647, 41.8010483012917]]
    )

    xpos_rec, ypos_rec = uc.error_bar_positions(xdat, sigx, ydat, sigy, rho)

    assert xpos_exp == pytest.approx(xpos_rec)
    assert ypos_exp == pytest.approx(ypos_rec)


def test_error_bar_position_values():
    """Test error bar position with Thomas' Macro values."""
    xdat = np.array([49.5263059304119, -19.8909690562228])
    sigx = np.array([116.127138038472, 108.88204516431])
    ydat = np.array([94.8320931109692, -24.8886147157681])
    sigy = np.array([137.018340778937, 124.423571979121])
    rho = np.array([0.429886405472074, 0.421261303003762])

    xpos_exp = np.array(
        [
            [-37.0849520308128, 136.137563891636],
            [-27.8300266053324, 126.882638466156],
            [-102.671239953765, 62.8893018413192],
            [-90.620923879426, 50.8389857669805],
        ]
    )
    ypos_exp = np.array(
        [
            [-31.507663087177, 221.171849309115],
            [147.863137920647, 41.8010483012917],
            [-138.042347260128, 88.2651178285919],
            [26.8555451027364, -76.6327745342726],
        ]
    )

    xpos_rec, ypos_rec = uc.error_bar_positions(xdat, sigx, ydat, sigy, rho)

    assert xpos_exp == pytest.approx(xpos_rec)
    assert ypos_exp == pytest.approx(ypos_rec)


@pytest.mark.parametrize("val", [1, 13.42])
def test_make_np_iterable(val):
    """Turn an individual value into a np.ndarray of length 1."""
    val_rec = uc._make_np_iterable(val)

    assert isinstance(val_rec, np.ndarray)
    assert len(val_rec) == 1


@pytest.mark.parametrize("val", [[1, 2], (13.42, 15), [23]])
def test_make_np_iterable_list(val):
    """Turn a list / similar into a np.ndarray of length 1."""
    val_rec = uc._make_np_iterable(val)

    assert isinstance(val_rec, np.ndarray)
    assert len(val_rec) == len(val)


def test_make_np_iterable_type_error():
    """Raise a type error if a string is passed."""
    with pytest.raises(TypeError):
        _ = uc._make_np_iterable("s")



"""Linear regression of data sets with correlated and uncorrelated uncertainties."""

from .ceresfit_linreg import LinReg

# Package information
__version__ = "0.2.1"
__all__ = ["LinReg"]

__title__ = "ceresfit"
__description__ = (
    "Linear regression of data sets with correlated and uncorrelated uncertainties."
    "Methodology follows the work of Mahon (1996) with correction for errors. For "
    "details, see the readme file on https://github.com/galactic-forensics/CEREsFit."
)

__uri__ = "https://github.com/galactic-forensics/CEREsFit"
__author__ = "Reto Trappitsch"
__email__ = "reto@galactic-forensics.space"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2023, Reto Trappitsch"



"""Test Konrecker delta function."""

import numpy as np
import pytest

from ceresfit.ceresfit_linreg import kron_delta


def test_kron_delta():
    """Ensure that 1 is returned if two indexes are the same, zero otherwise."""
    assert kron_delta(1, 1) == 1
    assert kron_delta(1, 2) == 0


def test_kron_delta_array():
    """Kronecker delta return for an array of indexes."""
    in1 = np.array([1, 2, 3, 4])
    in2 = np.array([1, 2, 4, 4])
    in2alt = [1, 2, 4, 4]  # as list
    out_exp = np.array([1, 1, 0, 1])
    np.testing.assert_equal(kron_delta(in1, in2), out_exp)
    np.testing.assert_equal(kron_delta(in1, in2alt), out_exp)


def test_kron_delta_value_error():
    """Raise ValueError if shapes mismatch."""
    arr1 = np.array([1, 2])
    arr2 = np.array([1])
    num = 3
    with pytest.raises(ValueError) as err_info:
        kron_delta(num, arr1)
    err_msg = err_info.value.args[0]
    assert err_msg == "The inputs must have the same shape."

    with pytest.raises(ValueError) as err_info:
        kron_delta(arr1, arr2)
    err_msg = err_info.value.args[0]
    assert err_msg == "The inputs must have the same shape."



"""Test confidence interval calculation."""

import numpy as np
import pytest
import scipy.stats as stats

from ceresfit import LinReg


# STEPHAN TEST DATASET FROM MACRO
def test_linreg_ci_calc(stephan_ci_data):
    """Compare CI calculation with Stephan Macro."""
    p_conf = 0.95
    (
        xdat,
        sigx,
        ydat,
        sigy,
        rho,
        x_ci_exp,
        y_ci_min_st,
        y_ci_max_st,
    ) = stephan_ci_data

    my_reg = LinReg(xdat, sigx, ydat, sigy, rho, autocalc=True)

    # Stephan uses 1 sigma error bands: mulitply with student t factor
    y_ci = x_ci_exp * my_reg.slope[0] + my_reg.intercept[0]
    zfac = stats.t.ppf(1 - (1 - p_conf) / 2.0, len(xdat) - 2)
    y_ci_min_exp = y_ci - np.abs(y_ci - y_ci_min_st) * zfac
    y_ci_max_exp = y_ci + np.abs(y_ci - y_ci_max_st) * zfac

    xrange = np.array([x_ci_exp.min(), x_ci_exp.max()])
    x_ci_rec, y_ci_min_rec, y_ci_max_rec = my_reg.confidence_intervals(
        p_conf=p_conf, xrange=xrange, bins=len(x_ci_exp)
    )

    assert x_ci_exp == pytest.approx(x_ci_rec)
    assert y_ci_min_exp == pytest.approx(y_ci_min_rec)
    assert y_ci_max_exp == pytest.approx(y_ci_max_rec)


def test_linreg_uncertainty_band_calc(stephan_ci_data):
    """Compare uncertainty band calculation with Stephan Macro."""
    (
        xdat,
        sigx,
        ydat,
        sigy,
        rho,
        x_ub_exp,
        y_ub_min_exp,
        y_ub_max_exp,
    ) = stephan_ci_data

    my_reg = LinReg(xdat, sigx, ydat, sigy, rho, autocalc=False)

    xrange = np.array([x_ub_exp.min(), x_ub_exp.max()])
    x_ub_rec, y_ub_min_rec, y_ub_max_rec = my_reg.uncertainty_band(
        xrange=xrange, bins=len(x_ub_exp)
    )

    assert x_ub_exp == pytest.approx(x_ub_rec)
    assert y_ub_min_exp == pytest.approx(y_ub_min_rec, rel=1e-4)
    assert y_ub_max_exp == pytest.approx(y_ub_max_rec, rel=1e-4)


def test_linreg_uncertainty_band_sigma(stephan_ci_data):
    """Ensure that 2 sigma band is twice as wide as 1 sigma band."""
    xdat, sigx, ydat, sigy, rho, _, _, _ = stephan_ci_data

    my_reg = LinReg(xdat, sigx, ydat, sigy, rho, autocalc=False)

    x_ub1, y_ub1_min, y_ub1_max = my_reg.uncertainty_band(sigma=1)
    x_ub2, y_ub2_min, y_ub2_max = my_reg.uncertainty_band(sigma=2)

    band_1sig = np.abs(y_ub1_max - y_ub1_min)
    band_2sig = np.abs(y_ub2_max - y_ub2_min)

    assert 2 * band_1sig == pytest.approx(band_2sig)



"""Test the Mahon Linear Regression with the data sets from Stephan and Trappitsch."""

from pathlib import Path
from typing import List

import numpy as np
import pytest

from ceresfit import LinReg


DATASETS = ["set1.csv", "set2.csv", "set3.csv", "set4.csv", "set5.csv"]
# absolute precision for comparison of non MSWD parameters
PRECISION_ABS = [1e-8, 1e-8, 1e-8, 1e-8, 1e-8]


def read_dataset(ds: Path) -> List[np.ndarray]:
    """Read a specific dataset and return the data.

    :param ds: Path to the dataset csv file.
    :return: xdata, xunc, ydata, yunc, rho, fix_point, parameters
        Parameters contain: Slope, Slope_unc, Intercept, Intercept_unc, MSWD
    """
    load_kwargs = {"delimiter": ",", "skiprows": 1}
    max_rows = [None, None, None, None, None, 1, 1, 3, 3, 3, 3, 3]
    cols_to_read = 12
    read_data = []
    for it in range(cols_to_read):
        read_data.append(
            np.loadtxt(ds, usecols=it, max_rows=max_rows[it], **load_kwargs)
        )

    # prepare return data
    ret_data = read_data[:5]
    ret_data.append(np.array([read_data[5], read_data[6]]))
    ret_data.append(np.array(read_data[7:]).transpose())
    return ret_data


@pytest.mark.parametrize("case", zip(DATASETS, PRECISION_ABS))  # noqa: B905
def test_linreg_cases_a(case, ds_path):
    """Comparpe cases with correlated uncertainties, no fixed point (sets a)."""
    ds = ds_path.joinpath(case[0])
    xdat, xunc, ydat, yunc, rho, _, params_exp = read_dataset(ds)

    ind_params = 0  # index of which parameters for given case

    reg = LinReg(xdat, xunc, ydat, yunc, rho=rho)
    assert params_exp[ind_params][:4] == pytest.approx(reg.parameters[:4], abs=case[1])
    assert params_exp[ind_params][4] == pytest.approx(reg.parameters[4], rel=1e-3)

    # check mswd and chi squared
    assert reg.chi_squared == pytest.approx(reg.mswd * (len(xdat) - 2))


@pytest.mark.parametrize("case", zip(DATASETS, PRECISION_ABS))  # noqa: B905
def test_linreg_cases_b(case, ds_path):
    """Comparpe cases with uncorrelated uncertainties, no fixed point (sets b)."""
    ds = ds_path.joinpath(case[0])
    xdat, xunc, ydat, yunc, _, _, params_exp = read_dataset(ds)

    ind_params = 1  # index of which parameters for given case

    reg = LinReg(xdat, xunc, ydat, yunc, rho=None)
    assert params_exp[ind_params][:4] == pytest.approx(reg.parameters[:4], abs=case[1])
    assert params_exp[ind_params][4] == pytest.approx(reg.parameters[4], rel=1e-3)


@pytest.mark.parametrize("case", zip(DATASETS, PRECISION_ABS))  # noqa: B905
def test_linreg_cases_c(case, ds_path):
    """Comparpe cases with correlated uncertainties, with fixed point (sets c)."""
    ds = ds_path.joinpath(case[0])
    xdat, xunc, ydat, yunc, rho, fixpt, params_exp = read_dataset(ds)

    ind_params = 2  # index of which parameters for given case

    reg = LinReg(xdat, xunc, ydat, yunc, rho=rho, fixpt=fixpt)
    assert params_exp[ind_params][:4] == pytest.approx(reg.parameters[:4], abs=case[1])
    assert params_exp[ind_params][4] == pytest.approx(reg.parameters[4], rel=1e-3)

    # check mswd and chi squared
    assert reg.chi_squared == pytest.approx(reg.mswd * (len(xdat) - 1))



"""Configuration file for testing the package with ``nox``."""

import nox


nox.options.sessions = "lint", "tests", "xdoctest"

package = "ceresfit"
locations = "ceresfit", "tests", "noxfile.py"
python_suite = ["3.11", "3.10", "3.9", "3.8"]
python_main = "3.10"


@nox.session(python=python_main)
def build(session):
    """Pack iniabu for release on PyPi."""
    session.install("flit")
    session.run("flit", "build")


@nox.session(python=python_main)
def lint(session):
    """Lint project using ``flake8``."""
    args = session.posargs or locations
    session.install("--upgrade", "pip")
    session.install(".[test]")
    session.run("flake8", *args)


@nox.session(python=python_suite)
def tests(session):
    """Test the project using ``pytest``."""
    session.install("--upgrade", "pip")
    session.install(".[test]")
    session.run("pytest")


@nox.session(python=python_main)
def safety(session):
    """Safety check for all dependencies."""
    session.install("--upgrade", "pip")
    session.install("safety")
    session.install(".[dev,test]")
    session.run(
        "safety",
        "check",
        "--full-report",
    )


@nox.session(python=python_main)
def xdoctest(session):
    """Test docstring examples with xdoctest."""
    args = session.posargs or ["all"]
    session.install("--upgrade", "pip")
    session.install("xdoctest[all]")
    session.install(".[dev,test]")
    session.run("python", "-m", "xdoctest", package, *args)



"""Uncertainty calculations that are outside the linear regression routine."""

from typing import Any, Tuple, Union

import numpy as np


def error_bar_angle_length(
    sigx: Union[float, np.ndarray],
    sigy: Union[float, np.ndarray],
    rho: Union[float, np.ndarray],
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate the angles and length of a given error bar.

    For a given uncertainty, calculate the angle and length of the rotated error bar
    dependant on the uncertainties and correlation coefficient. Returned are two arrays,
    an angle array and a length array. Both of them will have twice as many entries
    as the original data arrays. For each line in the data arrays, two angles and
    two lengths will be returned for the given error bars.

    :param sigx: X uncertainty
    :param sigy: Y uncertainty
    :param rho: Correlation factor

    :return: angles (radian), lengths
    """
    sigx = _make_np_iterable(sigx)
    sigy = _make_np_iterable(sigy)
    rho = _make_np_iterable(rho)
    if rho.shape != sigx.shape:
        rho = np.zeros_like(sigx) + rho

    sigxy = rho * sigx * sigy

    # matrix calculations
    trace_a = sigx**2 + sigy**2
    det_a = sigx**2 * sigy**2 - sigxy**2

    # eigenvalues
    lambda_1 = trace_a / 2 + np.sqrt((trace_a / 2) ** 2 - det_a)
    lambda_2 = trace_a / 2 - np.sqrt((trace_a / 2) ** 2 - det_a)

    # slopes of error bars
    m1 = (lambda_1 - sigx**2) / sigxy
    m2 = (lambda_2 - sigx**2) / sigxy

    # angles
    alpha1 = np.arctan(m1)
    alpha2 = np.arctan(m2)

    # sort the return values
    alphas_ret = np.zeros(2 * len(sigx))
    lengths_ret = np.zeros(2 * len(sigx))

    for it in range(len(alpha1)):
        alphas_ret[2 * it] = alpha1[it]
        alphas_ret[2 * it + 1] = alpha2[it]
        lengths_ret[2 * it] = np.sqrt(lambda_1[it])
        lengths_ret[2 * it + 1] = np.sqrt(lambda_2[it])

    return alphas_ret, lengths_ret


def error_bar_positions(
    xdat: Union[float, np.ndarray],
    sigx: Union[float, np.ndarray],
    ydat: Union[float, np.ndarray],
    sigy: Union[float, np.ndarray],
    rho: Union[float, np.ndarray],
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate the positions of rotated error bars.

    This routine returns two arrays, both are twice as long as the initial data with two
    positions per line. Two arrays are returned, one with x positions and one with
    y positions. The first line in each array corresponds to the corresponding x
    and y coordinates between which a line must be drawn to represent an error bar
    for the first datapoint. The second entry represents the second error bar for the
    first entry, etc.

    :param xdat: X data
    :param sigx: X uncertainty
    :param ydat: Y data
    :param sigy: Y uncertainty
    :param rho: Correlation factor

    :return: x positions, y positions
    """
    angles, lengths = error_bar_angle_length(sigx, sigy, rho)

    # calculate deltas for x and y
    dx = np.cos(angles) * lengths
    dy = np.sin(angles) * lengths

    xret = np.zeros((len(dx), 2))
    yret = np.zeros_like(xret)

    # write the data point values into xret and yret
    if np.asarray(xdat).shape == ():
        xret += xdat
        yret += ydat
    else:
        for it in range(len(xdat)):
            xret[2 * it] += xdat[it]
            xret[2 * it + 1] += xdat[it]
            yret[2 * it] += ydat[it]
            yret[2 * it + 1] += ydat[it]

    # add / subtract the dx, dy values
    xret[:, 0] -= dx
    xret[:, 1] += dx
    yret[:, 0] -= dy
    yret[:, 1] += dy

    return xret, yret


def _make_np_iterable(value: Any) -> np.ndarray:
    """Turn a value into an np.ndarray.

    :param value: Value(s) to turn into an array.

    :return: Given value as a numpy array.

    :raises TypeError: Unsupported type for value.
    """
    if isinstance(value, str):
        raise TypeError(f"Unsupported type {type(value)}.")

    if hasattr(value, "__iter__"):
        return np.asarray(value)
    else:  # it's a single value -> turn into array of length 1
        return np.asarray(value).reshape(1)
